/*
 * $Id: debug.c,v 1.1.1.1 2000/06/22 08:04:48 matley Exp $
 */
/*
 * Copyright (c) 2000 Massachusetts Institute of Technology
 * Copyright (c) 2000 Matteo Frigo
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <stdarg.h>
#include <stdio.h>
#include <cilk.h>
#include <cilk-internal.h>

/***********************************************************\
 * Debugging & tracing
\***********************************************************/

void Cilk_dprintf(CilkWorkerState *const ws, const char *fmt,...)
{
     va_list l;

     /* ensure nice interleaving of debugging messages */
     Cilk_mutex_wait(ws, &USE_SHARED(dprintf_lock));

     if (ws >= 0)
	  fprintf(stderr, "%d: ", ws->self);
     else
	  fprintf(stderr, "global: ");

     va_start(l, fmt);
     vfprintf(stderr, fmt, l);
     fflush(stderr);
     va_end(l);
     Cilk_mutex_signal(&USE_SHARED(dprintf_lock));
}

void Cilk_die_internal(CilkWorkerState *const ws, const char *fmt,...)
{
     va_list l;

     Cilk_mutex_wait(ws, &USE_SHARED(dprintf_lock));

     if (ws)
	  fprintf(stderr, "%d: fatal error: ", ws->self);
     else
	  fprintf(stderr, "global: fatal error: ");

     va_start(l, fmt);
     vfprintf(stderr, fmt, l);
     va_end(l);
     Cilk_mutex_signal(&USE_SHARED(dprintf_lock));

     /* ensure only the first fatal error produces death of the program */
     if (USE_PARAMETER(dump_core) && Cilk_mutex_try(&USE_SHARED(die_lock)))
	  abort();		/* dump core */
     else
          exit(1);              /* don't dump core */
}

void Cilk_die(const char *fmt,...)
{
     va_list l;

     Cilk_mutex_wait(NULL, &USE_SHARED(dprintf_lock));

     fprintf(stderr, "fatal error: ");

     va_start(l, fmt);
     vfprintf(stderr, fmt, l);
     va_end(l);
     Cilk_mutex_signal(&USE_SHARED(dprintf_lock));

     /* ensure only the first fatal error produces death of the program */
     if (USE_PARAMETER(dump_core) && Cilk_mutex_try(&USE_SHARED(die_lock)))
	  abort();		/* dump core */
     else
          exit(1);              /* don't dump core */
}

void Cilk_debug_init(void)
{
     Cilk_mutex_init(&USE_SHARED(dprintf_lock));
     Cilk_mutex_init(&USE_SHARED(die_lock));
}
