/*
 * $Id: invoke-main.c,v 1.1.1.1 2000/06/22 08:04:49 matley Exp $
 */

/*
 * Copyright (c) 2000 Massachusetts Institute of Technology
 * Copyright (c) 2000 Matteo Frigo
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/* startup routines. You are not expected to understand this file */

#include <stdio.h>
#include <cilk.h>
#include <cilk-internal.h>
#include <cilk-cilk2c-pre.h>
#include <cilk-cilk2c.h>

/************************************************************
 *  Initial thread
 ************************************************************/

/* This is a hand-compiled procedure that calls cilk_main */
struct invoke_main_frame {
     CilkStackFrame header;
     int argc;
     char **argv;
     int (*main)(CilkWorkerState *const ws, int argc, char *argv[]);
     int retval;
     int res; 
};


struct invoke_main_catch_inlet_args {int res;};

/* 
 * This inlet is created for the purpose of Cilk_exit(). 
 * It sets retval to the exit value, and calls abort.
 * So, when this inlet is excuted, the user's cilk_main
 * should be aborted as a child of invoke_main.
 */
static void 
invoke_main_catch_inlet(CilkWorkerState *const _cilk_ws,
			struct invoke_main_frame *_cilk_frame,
			struct invoke_main_catch_inlet_args *inletargs,
			void* inletresult)
{
     USE_SHARED(retval) = inletargs->res;
     CILK2C_ABORT_STANDALONE();
}

static void invoke_main_slow(CilkWorkerState *const _cilk_ws,
			     struct invoke_main_frame *_cilk_frame)
{
     int argc;
     char **argv;
     int tmp;

     CILK2C_START_THREAD_SLOW();
     switch (_cilk_frame->header.entry) {
     case 1: 
	  goto _sync1;
     case 2: 
	  goto _sync2;
     }
     
     _cilk_ws->user_work = 0;
     _cilk_ws->user_critical_path = 0;
     WHEN_CILK_TIMING(_cilk_frame->header.cp = (Cilk_time) 0);
     WHEN_CILK_TIMING(_cilk_frame->header.work = (Cilk_time) 0);
     WHEN_CILK_TIMING(_cilk_frame->header.mycp = (Cilk_time) 0);
     WHEN_CILK_TIMING(_cilk_ws->last_cp_time = Cilk_get_time());

     argc = _cilk_frame->argc;
     argv = _cilk_frame->argv;

     _cilk_frame->header.index = 
	  (char *) &_cilk_frame->retval - (char *) _cilk_frame;
     _cilk_frame->header.entry=1;
     CILK2C_BEFORE_SPAWN_SLOW();
     CILK2C_PUSH_FRAME(_cilk_frame);
     tmp = _cilk_frame->main(_cilk_ws, argc, argv);
     CILK2C_XPOP_FRAME_RESULT(_cilk_frame,/* return nothing */,tmp);
     CILK2C_AFTER_SPAWN_SLOW();

     USE_SHARED(retval) = tmp;

     if (0) {
     _sync1:
	  ;
     }
     CILK2C_AT_THREAD_BOUNDARY_SLOW();
     
     CILK2C_BEFORE_SYNC_SLOW();
     _cilk_frame->header.entry=2;
     if (CILK2C_SYNC) {
	  return;
     _sync2:
	  ;
     }
     CILK2C_AFTER_SYNC_SLOW();
     CILK2C_AT_THREAD_BOUNDARY_SLOW();

     USE_SHARED(done) = 1;

     WHEN_CILK_TIMING(USE_SHARED(critical_path) = _cilk_frame->header.mycp);
     WHEN_CILK_TIMING(USE_SHARED(total_work) = _cilk_frame->header.work);

     Cilk_remove_and_free_closure_and_frame(_cilk_ws,
					    &_cilk_frame->header, 
					    _cilk_ws->self);

     return;
}


static CilkProcInfo invoke_main_sig[]=
{
     {sizeof(int), sizeof(struct invoke_main_frame), invoke_main_slow, 0, 0},
     {sizeof(int), 0, invoke_main_catch_inlet,
      sizeof(struct invoke_main_catch_inlet_args ),
      CILK_OFFSETOF(struct invoke_main_frame, res)},
     {0, 0, 0, 0, 0}
};


Closure *Cilk_create_initial_thread(
     int (*main)(CilkWorkerState *const ws, int argc, char *argv[]),
     int argc, char *argv[])
{
     Closure *t;
     struct invoke_main_frame *f;

     /* create a frame for invoke_cilk_main */
     t = Cilk_Closure_create_malloc(NULL);
     t->parent = (Closure *) NULL;
     t->join_counter = 0;
     t->status = CLOSURE_READY;

     f = Cilk_malloc(sizeof(struct invoke_main_frame));
     f->header.entry = 0;
     f->header.sig = invoke_main_sig;
     WHEN_CILK_DEBUG(f->header.magic = CILK_STACKFRAME_MAGIC);

     f->argc = argc;
     f->argv = argv;
     f->main = main;

     t->frame = &f->header;
     return t;
}

/*************************************************************
 * Exit-related routines
 *************************************************************/
struct _Cilk_really_exit_frame {
     CilkStackFrame header;
     struct {
	  int val;
     } scope0;
};

static CilkProcInfo _Cilk_really_exit_sig[] =
{
     {0, sizeof(struct _Cilk_really_exit_frame), /* no slow version */ 0, 0, 0}
};

void Cilk_really_exit(CilkWorkerState *const _cilk_ws, int val)
{
     struct _Cilk_really_exit_frame *_cilk_frame;
     CILK2C_INIT_FRAME(_cilk_frame, sizeof(struct _Cilk_really_exit_frame), 
		       _Cilk_really_exit_sig);
     CILK2C_START_THREAD_FAST();

     {
	  Cilk_really_exit_1(_cilk_ws, val);
	  {
	       CILK2C_BEFORE_RETURN_FAST();
	       return;
	  }
     }
}
